/*	$Id$	*/

/*
 * Copyright (c) 2012, Internet Initiative Japan, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <libarms.h>
#include <libarms_resource.h>
#include <axp_extern.h>

#include <libarms_log.h>
#include <arms_xml_tag.h>
#include <module_db_mi.h>
#include <libarms/base64.h>
#include <libarms/malloc.h>
#include <transaction/transaction.h>
#include <protocol/arms_methods.h>

#include "compat.h"

/*
 * Callback Functions
 */
/* context alloc */
static void *read_storage_context(tr_ctx_t *tr_ctx);
/* copy argument */
static int
read_storage_cparg(AXP *axp, uint32_t pm_type, int tag, const char *buf, size_t len, tr_ctx_t *tr_ctx);
/* done */
static int
read_storage_done(transaction *, char *buf, int, int *);
/* exec */
static int
read_storage_exec(transaction *tr);
/* context free */
static void read_storage_release(tr_ctx_t *tr_ctx);
/* revertdone */
static void read_storage_revertdone(transaction *);

/*
 * XML Schema: read-storage-start-request
 */
static struct axp_schema arms_push_readstorage_req[] = {
	{ARMS_TAG_STORAGE, "storage", AXP_TYPE_TEXT,
	 NULL, push_default_hook, NULL},
	{0, NULL, 0, NULL, NULL, NULL}
};
struct axp_schema read_storage_start_req = {
	ARMS_TAG_READSTORAGE_SREQ, "read-storage-start-request", AXP_TYPE_CHILD,
	 	NULL, push_default_hook, arms_push_readstorage_req
};

/*
 * Method definition
 */
arms_method_t read_storage_methods = {
	.pm_type       = ARMS_TR_READ_STORAGE,
	.pm_string     = "read-storage",
	.pm_schema     = &read_storage_start_req,
	.pm_flags      = 0,
	.pm_response   = build_generic_res,
	.pm_done       = read_storage_done,
	.pm_exec       = read_storage_exec,
	.pm_copyarg    = read_storage_cparg,
	.pm_rollback   = NULL,
	.pm_context    = read_storage_context,
	.pm_release    = read_storage_release,
	.pm_revertdone = read_storage_revertdone,
};

struct read_storage_result {
	int rv;
	char result[1024];
	char term;	/* trailing NUL */
	int blen;
	int done_flg;
	SIMPLEQ_ENTRY(read_storage_result) next;
};

/*
 * Method implementations
 */

#define BEGIN        1
#define FIRST_RESULT 2
#define NEXT_RESULT  3
#define DONE_RESULT  4
#define DONE         5

struct read_storage_args {
	int props_id; /* storage type: running, candidata, or backup */
	int mod_index;
	int mod_max;
	uint32_t mod_id;
	int permanent;
	int next;
	int state;
	SIMPLEQ_HEAD(, read_storage_result) cb_result;
	struct read_storage_result *cbr_next;
};

/*
 * Context Alloc
 */
static void *
read_storage_context(tr_ctx_t *tr_ctx)
{
	struct read_storage_args *arg;
	arms_context_t *res = arms_get_context();

	if (res->callbacks.read_config_cb == NULL) {
		tr_ctx->result = 505;
		return 0;
	}

	arg = CALLOC(1, sizeof(*arg));
	if (arg != NULL) {
		arg->state = BEGIN;
	} else {
		tr_ctx->result = 413; /*Resource Exhausted*/
	}
	SIMPLEQ_INIT(&arg->cb_result);

	return arg;
}

/*
 * Context Free
 */
static void
read_storage_release(tr_ctx_t *tr_ctx)
{
	struct read_storage_args *arg;
	struct read_storage_result *cbr;

	arg = tr_ctx->arg;
	if (arg) {
		while ((cbr = SIMPLEQ_FIRST(&arg->cb_result)) != NULL) {
			SIMPLEQ_REMOVE(&arg->cb_result, cbr,
			    read_storage_result, next);
			FREE(cbr);
		}

		FREE(tr_ctx->arg);
	}
}

/*
 * Copy argument
 */
static int
parse_propsid(const char *buf, size_t len)
{
	if (buf == NULL)
		return -1;
	if (len <= 0)
		return -1;

#if 0 /* libarms doesn't support startup */
	if (strncmp("startup", buf, len) == 0) {
		return ARMS_CONFIG_STARTUP;
	}
#endif
	else if (strncmp("candidate", buf, len) == 0) {
		return ARMS_CONFIG_CANDIDATE;
	}
	else if (strncmp("running", buf, len) == 0) {
		return ARMS_CONFIG_RUNNING;
	}
	else if (strncmp("backup", buf, len) == 0) {
		return ARMS_CONFIG_BACKUP;
	}

	return -1;
}

static int
read_storage_cparg(AXP *axp, uint32_t pm_type, int tag,
		   const char *buf, size_t len, tr_ctx_t *tr_ctx)
{
	struct read_storage_args *arg = tr_ctx->arg;

	if (tag == ARMS_TAG_STORAGE) {
		arg->props_id = parse_propsid(buf, len);
		if (arg->props_id < 0) {
			/* storage type is invalid. */
			tr_ctx->result = 203;
		}
	} else {
		/* other tag (error?) */
	}
	return 0;
}

/*
 * read-storage-done
 *
 * like read-storage-done, but <md-config> tag has no result attribute.
 *  so fat, empty config means error.  umm...
 *
 * SMF SDK 1.00 (and 1.10) cannot receive error result via any tag.
 * - md-config has no result attribute.
 * - transaction-result doesn't work between proxy and RS.
 * workaround: done-req doesn't include md-config tag means error.
 */
static int
read_storage_done(transaction *tr, char *buf, int len, int *wrote)
{
	tr_ctx_t *tr_ctx = &tr->tr_ctx;
	struct read_storage_args *arg = tr_ctx->arg;
	struct read_storage_result *cbr;

	switch (arg->state) {
	case BEGIN:
		libarms_log(ARMS_LOG_DEBUG, "Generate read-storage-done");
		*wrote = arms_write_begin_message(tr, buf, len);
		arg->cbr_next = SIMPLEQ_FIRST(&arg->cb_result);

		if (tr_ctx->result == 100)
			arg->state = FIRST_RESULT;
		else
			arg->state = DONE;
		return TR_WANT_WRITE;
	case FIRST_RESULT:
		/*
		 * mod_index++ if err or FINISHED.
		 */
		arms_get_module_id(&arg->mod_id, &arg->permanent, arg->mod_index);
		cbr = arg->cbr_next;
		if (cbr == NULL) {
			if (arg->permanent == ARMS_PERMANENT_INVALID) {
				*wrote = snprintf(buf, len,
						"<md-config id=\"%d\">",
						arg->mod_id);
			} else {
				*wrote = snprintf(buf, len,
						"<md-config id=\"%d\" "
						"permanent=\"%d\" >",
						arg->mod_id, arg->permanent);
			}
			arg->state = DONE_RESULT;
			return TR_WANT_WRITE;
		} else {
			arg->cbr_next = SIMPLEQ_NEXT(cbr, next);
		}

		if (ARMS_RESULT_IS_BYTES(cbr->rv)) {
			/* binary */
			if (arg->permanent == ARMS_PERMANENT_INVALID) {
				*wrote = snprintf(buf, len,
						"<md-config id=\"%d\" "
						"encoding=\"base64\">",
						arg->mod_id);
			} else {
				*wrote = snprintf(buf, len,
						"<md-config id=\"%d\" "
						"permanent=\"%d\" "
						"encoding=\"base64\">",
						arg->mod_id, arg->permanent);
			}
			*wrote += arms_base64_encode(buf + *wrote, len - *wrote,
				   cbr->result, cbr->blen);
		} else {
			/* text or error */
			if (arg->permanent == ARMS_PERMANENT_INVALID) {
				*wrote = snprintf(buf, len,
						"<md-config id=\"%d\">%s",
						arg->mod_id,
						arms_escape(cbr->result));
			} else {
				*wrote = snprintf(buf, len,
						"<md-config id=\"%d\" "
						"permanent=\"%d\">%s",
						arg->mod_id, arg->permanent,
						arms_escape(cbr->result));
			}
		}
		if (cbr->done_flg) {
			arg->state = DONE_RESULT;
		} else {
			arg->state = NEXT_RESULT;
		}
		return TR_WANT_WRITE;
	case NEXT_RESULT:
		cbr = arg->cbr_next;
		if (cbr == NULL) {
			arg->state = DONE_RESULT;
			*wrote = 0;
			return TR_WANT_WRITE;
		} else {
			arg->cbr_next = SIMPLEQ_NEXT(cbr, next);
		}
		if (ARMS_RESULT_IS_ERROR(cbr->rv)) {
			arg->state = DONE_RESULT;
			*wrote = 0;
		} else if (ARMS_RESULT_IS_BYTES(cbr->rv)) {
			*wrote = arms_base64_encode(buf, len, cbr->result, cbr->blen);
		} else {
			*wrote = strlcpy(buf, arms_escape(cbr->result), len);
		}
		if (cbr->done_flg)
			arg->state = DONE_RESULT;
		return TR_WANT_WRITE;
	case DONE_RESULT:
		*wrote = snprintf(buf, len, "</md-config>");
		arg->mod_index++;
		if (arg->mod_index >= arg->mod_max)
			arg->state = DONE;
		else
			arg->state = FIRST_RESULT;
		return TR_WANT_WRITE;
	case DONE:
		*wrote = arms_write_end_message(tr, buf, len);
		libarms_log(ARMS_LOG_DEBUG, "Read Storage Execute done.");
		return TR_WRITE_DONE;
	default:
		break;
	}
	return TR_FATAL_ERROR;
}

static int
read_storage_exec(transaction *tr)
{
	tr_ctx_t *tr_ctx = &tr->tr_ctx;
	arms_context_t *res = arms_get_context();
	struct read_storage_args *arg = tr_ctx->arg;
	struct read_storage_result *cbr;
	int remain_blen = 0;
	char remain_bdata[2];
	int i;

	arg->mod_max = arms_count_module();
	for (i = 0; i < arg->mod_max; i++) {
		arg->next = ARMS_FRAG_FIRST;
		remain_blen = 0;
		if (arms_get_module_id(&arg->mod_id, &arg->permanent, i) != 0) {
			tr_ctx->result = 402; /*402 SA Failure*/
			return 0;
		}
		do {
			cbr = CALLOC(1, sizeof(*cbr));
			if (!cbr) {
				tr_ctx->result = 413;/*413 Resource Exhausted*/
				return 0;
			}
			SIMPLEQ_INSERT_TAIL(&arg->cb_result, cbr, next);

			if (remain_blen) {
				memcpy(cbr->result, remain_bdata, remain_blen);
			}
			cbr->rv = res->callbacks.read_config_cb(
					arg->mod_id,
					arg->props_id,
					cbr->result + remain_blen,
					sizeof(cbr->result) - remain_blen,
					&arg->next,
					res->udata);

			if (ARMS_RESULT_IS_BYTES(cbr->rv)) {
				/* binary */
				cbr->blen = ROUND_BASE64_BINARY(ARMS_RV_DATA_MASK(cbr->rv) + remain_blen);
				remain_blen = ARMS_RV_DATA_MASK(cbr->rv) + remain_blen - cbr->blen;
				memcpy(remain_bdata, &cbr->result[cbr->blen], remain_blen);
			} else {
				/* text or error */
				remain_blen = 0;
				if (ARMS_RESULT_IS_ERROR(cbr->rv))
					break;
			}
			if ((arg->next & ARMS_FRAG_FINISHED))
				break;
			arg->next = ARMS_FRAG_CONTINUE;
		} while (1);

		if (remain_blen) {
			cbr = CALLOC(1, sizeof(*cbr));
			if (!cbr) {
				tr_ctx->result = 413;/*413 Resource Exhausted*/
				return 0;
			}
			SIMPLEQ_INSERT_TAIL(&arg->cb_result, cbr, next);
			cbr->rv = ARMS_RESULT_BYTES(remain_blen);
			memcpy(cbr->result, remain_bdata, remain_blen);
			cbr->blen = remain_blen;
		}
		cbr->done_flg++;
	};

	return 0;
}

static void
read_storage_revertdone(transaction *tr)
{
	tr_ctx_t *tr_ctx = &tr->tr_ctx;
	struct read_storage_args *arg = tr_ctx->arg;

	arg->state = BEGIN;
}
